from pathlib import Path
import json

def train_model(df_train,df_eval, config, format_df_for_pet_pattern,dataframe_to_jsonl):
    train_pet_df = format_df_for_pet_pattern(df_train, ['input_description', 'genereted_conditions'], 'label')
    val_pet_df = format_df_for_pet_pattern(df_eval, ['input_description', 'genereted_conditions'], 'label')
    
    Path("data/szinternal").mkdir(parents=True, exist_ok=True)
    dataframe_to_jsonl(train_pet_df, 'data/szinternal/train.jsonl')
    dataframe_to_jsonl(val_pet_df, 'data/szinternal/val.jsonl')
    return config


def get_the_model(output_path,Config,load_model):

  model_path = output_path + "/best_model.pt"
  config_path = output_path + "/config.json"

  config = json.load(open(config_path))
  config =  Config(kwargs=config)

  model =  load_model(config,model_path)
  return model


from model_utils import get_the_model

from sz_metrics import SZ_Calibration_Metric
import os
import torch
import pandas as pd

def get_performance_and_ue_score(df_test,output_path,Config,load_model,batch_data_from_dataframe):
  batches = batch_data_from_dataframe(df_test, desc_cols=['input_description', 'genereted_conditions'], batch_size=4)

  adapet_logits = []
  adapet_pred = []

  classifier_logits = []
  classifier_pred = []
  model = get_the_model(output_path,Config,load_model)

  with torch.no_grad():
    for batch in batches:

      pred_lbl, lbl_logits, cls_prob = model.predict(batch) # make it editable
      cls_pred = torch.argmax(cls_prob, dim=1)

      adapet_logits.extend(lbl_logits.cpu().numpy().tolist())
      adapet_pred.extend(pred_lbl.cpu().numpy().tolist())

      classifier_logits.extend(cls_prob.cpu().numpy().tolist())
      classifier_pred.extend(cls_pred.cpu().numpy().tolist())

  true_label = [1 if x == 'yes' else 0 for x in df_test["label"]]

  df_test['adapet_machine_label'] = ['yes' if x == 1 else "no" for x in adapet_pred]
  df_test["adapet_logits"] = adapet_logits

  df_test['calssifier_machine_label'] = ['yes' if x == 1 else "no" for x in classifier_pred]
  df_test["classifier_logits"] = classifier_logits

  

  adapet_pscore_dict = SZ_Calibration_Metric.get_performance_metrics(true_label,adapet_pred)
  classifier_pscore_dict = SZ_Calibration_Metric.get_performance_metrics(true_label,classifier_pred)

  adapet_ue_score_dict = SZ_Calibration_Metric.get_ue_metrics(true_label ,adapet_logits)
  classifier_ue_score_dict = SZ_Calibration_Metric.get_ue_metrics(true_label ,classifier_logits)

  return adapet_pscore_dict,classifier_pscore_dict,adapet_ue_score_dict,classifier_ue_score_dict, df_test


def save_preds(df_test,output_path,sg_name,num_examples,Config,load_model,batch_data_from_dataframe):

    adapet_pscore_dict,classifier_pscore_dict,adapet_ue_score_dict,classifier_ue_score_dict, df_test = get_performance_and_ue_score(df_test,output_path,Config,load_model,batch_data_from_dataframe)
    # link the label and probability into a dataframe
    adapet_score_dict = {**adapet_pscore_dict, **adapet_ue_score_dict}
    classifier_score_dict = {**classifier_pscore_dict, **classifier_ue_score_dict}

    adapet_pscore_dict_df = pd.DataFrame([adapet_score_dict])
    adapet_pscore_dict_df.insert(loc=0, column='Name', value= sg_name +"_"+ str(num_examples) + "_"+ 'adapet_score')

    classifier_pscore_dict_df = pd.DataFrame([classifier_score_dict])
    classifier_pscore_dict_df.insert(loc=0, column='Name', value= sg_name + "_" + str(num_examples) + "_" + 'classifier_score')

    df = pd.concat([ adapet_pscore_dict_df, classifier_pscore_dict_df], axis=0)

    save_preds_path = os.path.exists('/content/adapet/save_preds')
    if not save_preds_path:
      os.mkdir('/content/adapet/save_preds')
    df.to_csv("/content/adapet/save_preds/" + sg_name +"_"+ str(num_examples) + ".csv",index=False)

    save_test_data_path = os.path.exists('/content/adapet/save_test_data/')
    if not save_test_data_path:
      os.mkdir('/content/adapet/save_test_data/')
    df_test.to_csv("/content/adapet/save_test_data/" + sg_name +"_"+ str(num_examples) + ".csv",index=False)


import pandas as pd
import numpy as np

def test_set_select(df, seed = 42):
  train, validate, test = np.split(df.sample(frac=1, random_state= seed ),[int(.8*len(df)), int(.9*len(df))])

  df_yes = train.loc[train['label'] == 'yes'] 
  df_no = train.loc[train['label'] == 'no']

  if len(df_yes) > len(df_no):
    rem_ex = len(df_yes) - len(df_no)
    get_more_ex = df_yes.sample(rem_ex)
    train = pd.concat([train, get_more_ex, get_more_ex]).drop_duplicates(keep=False)
    test = pd.concat([test, get_more_ex], axis = 0)
  
  if len(df_no) > len(df_yes):
    rem_ex = len(df_no) - len(df_yes)
    get_more_ex = df_no.sample(rem_ex)
    train = pd.concat([train, get_more_ex, get_more_ex]).drop_duplicates(keep=False)
    test = pd.concat([test, get_more_ex], axis = 0)

  return train.reset_index(drop = True), validate.reset_index(drop = True), test.reset_index(drop = True)

def select_num_examples(train_df, num_examples):
  print("num_examples" , num_examples, "\n")

  df_yes = train_df.loc[train_df['label'] == 'yes'] 
  df_no = train_df.loc[train_df['label'] == 'no'] 
  print("num_yes : " ,len(df_yes)," num_no : ", len(df_no), "\n")

  if num_examples == 'ALL':
    train_df = train_df

  # elif len(df_yes) != len(df_no):
  #   train_df = train_df

  else:
    train_df = pd.concat([df_yes.sample(int(num_examples/2)),df_no.sample(int(num_examples/2))], axis = 0)
  
  return train_df.reset_index(drop = True)



import torch
import os

from adapet_helper_function import *
from training_ex_selection import *
from save_score import *
from model_utils import *
from collections.abc import Mapping

def start_training(df, segment_name, seed, checkpoint_path,config_dict, function_dict : dict):

  
    
    if isinstance(function_dict, Mapping) == False:
        raise ("Dictionary of functions is missing")

    if 'config' in vars():
        del(config)
    if 'model' in vars():
        del(model)
    if 'tokenizer' in vars():
        del(tokenizer)
    if 'checkpoint' in vars():
        del(checkpoint)
    torch.cuda.empty_cache()

    train_data, val_df, test_df = test_set_select(df,seed) # seed value? Done---

    # segment_name_dir = segment_name[:-4] ## /content/adapet/datasets/Consumer_Electronics.csv >> /content/adapet/datasets/Consumer_Electronics ## remove .csv from end
    sg_name = segment_name[25:-4] ## removing '/content/adapet/datasets/' and '.csv' to get the dataset name
    
    num_example_li = [10, 20, 30, 40, 50, 60, 70, 100, 'ALL'] # 27 yes = 13 no = 14 ## test this function...? no idea..

    for num_examples in num_example_li: # break the loop num_e > len(df) Done

      if  num_examples > len(train_data):
        break

      df_train = select_num_examples(train_data, num_examples)  ## check test this function...? Done
      print("training data : ", len(df_train))
    
      config = config_dict
      os.environ['SZ_ROOT'] = ''    

      output_path = os.path.join(os.environ['SZ_ROOT'], "exp_out", sg_name + '_' + str(num_examples))
      if not os.path.exists(output_path):
          os.makedirs(output_path)
      
      print(output_path)

      config['exp_dir'] = output_path
      config = train_model(df_train, val_df ,config, function_dict["format_df_for_pet_pattern"],function_dict["dataframe_to_jsonl"])
      config = function_dict['Config'](kwargs=config, mkdir=False)
      train(config, output_path, checkpoint_path,
            function_dict["Batcher"],
            function_dict["adapet"],
            function_dict["device"],
            function_dict["update_dict_val_store"],
            function_dict["get_avg_dict_val_store"],
            function_dict["dev_eval"]
            )
      
      save_preds(test_df,output_path,sg_name,num_examples,function_dict["Config"],function_dict["load_model"],function_dict["batch_data_from_dataframe"])


import transformers as tfrms
from tqdm import tqdm
import torch 

from torch.optim.lr_scheduler import LambdaLR


def get_linear_schedule_with_warmup(optimizer, num_warmup_steps, num_training_steps, last_epoch=-1):
    """
    Create a schedule with a learning rate that decreases linearly from the initial lr set in the optimizer to 0,
    after a warmup period during which it increases linearly from 0 to the initial lr set in the optimizer.
    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The total number of training steps.
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.
    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
        )

    return LambdaLR(optimizer, lr_lambda, last_epoch)

def train(config , outputpath, checkpoint_path,Batcher,adapet,device,update_dict_val_store,get_avg_dict_val_store,dev_eval):
    '''
    Trains the model
    :param config:
    :return:
    '''

    tokenizer = tfrms.AutoTokenizer.from_pretrained(config.pretrained_weight)
    batcher = Batcher(config, tokenizer, config.dataset)
    dataset_reader = batcher.get_dataset_reader()
    print('Initializing Model\n')
    

    model = adapet(config, tokenizer, dataset_reader)
    checkpoint = torch.load(checkpoint_path)
    model.load_state_dict(checkpoint)
    model.to(device)
    print(type(model))
    print("------------------------------------------------------")
    ### Create Optimizer
    # Ignore weight decay for certain parameters
    no_decay_param = ['bias', 'LayerNorm.weight']
    optimizer_grouped_parameters = [
        {'params': [p for n, p in model.model.named_parameters() if not any(nd in n for nd in no_decay_param)],
         'weight_decay': config.weight_decay,
         'lr': config.lr},
        {'params': [p for n, p in model.model.named_parameters() if any(nd in n for nd in no_decay_param)],
         'weight_decay': 0.0,
         'lr': config.lr},
    ]
    optimizer = torch.optim.AdamW(optimizer_grouped_parameters, eps=1e-8)

    best_dev_acc = 0
    train_iter = batcher.get_train_batch()
    dict_val_store = None

    # Number of batches is assuming grad_accumulation_factor forms one batch
    tot_num_batches = config.num_batches * config.grad_accumulation_factor

    # Warmup steps and total steps are based on batches, not epochs
    num_warmup_steps = config.num_batches * config.warmup_ratio
    scheduler = get_linear_schedule_with_warmup(optimizer, num_warmup_steps, config.num_batches)

    for i in tqdm(range(tot_num_batches), desc='Train Batches'):
        # Get true batch_idx
        batch_idx = (i // config.grad_accumulation_factor)

        model.train()
        sup_batch = next(train_iter)
        loss, dict_val_update = model(sup_batch)
        loss = loss / config.grad_accumulation_factor
        loss.backward()

        if (i+1) % config.grad_accumulation_factor == 0:
            torch.nn.utils.clip_grad_norm_(model.parameters(), config.grad_clip_norm)
            optimizer.step()
            optimizer.zero_grad()
            scheduler.step()
      
        dict_val_store = update_dict_val_store(dict_val_store, dict_val_update, config.grad_accumulation_factor)
        print("Finished %d batches" % batch_idx, end='\r')
        
        if (batch_idx + 1) % config.eval_every == 0 and i % config.grad_accumulation_factor == 0:
            dict_avg_val = get_avg_dict_val_store(dict_val_store, config.eval_every)
            dict_val_store = None
            dev_acc, dev_logits = dev_eval(config, model, batcher, batch_idx, dict_avg_val)

            print("Global Step: %d Acc: %.3f" % (batch_idx, dev_acc) + '\n')
            if dev_acc > best_dev_acc:
                best_dev_acc = dev_acc
                name = outputpath + "/best_model.pt"
                torch.save(model.state_dict(),name)
                